import json
import logging
from pathlib import Path
from typing import Any, Dict, Iterator, List
from uuid import UUID

from pydantic import BaseModel

from cado.core.cell import Cell
from cado.core.cell_status import CellStatus
from cado.core.language import Language

logger = logging.getLogger(__name__)

CURRENT_VERSION = "0.1"


class Notebook(BaseModel):
    name: str
    version: str = CURRENT_VERSION
    cells: List[Cell] = []

    def update_cell_output_name(self, cell_id: UUID, output_name: str) -> None:
        """Set a notebook cell's output name.

        Args:
            cell_id (UUID): ID of the cell.
            output_name (str): Name for the output variable.
        """
        self.clear_cell(cell_id)

        output_names = set()
        for cell in self.cells:
            if cell.id != cell_id and cell.output_name != "":
                output_names.add(cell.output_name)

        cell = self.get_cell(cell_id)

        for child in self._get_children(cell):
            child.clear()

        if output_name in output_names:
            cell.set_error()
            cell.output_name = ""
            raise ValueError(f"Cell with output name \"{output_name}\" already exists in the notebook")

        cell.output_name = output_name

    def update_cell_input_names(self, cell_id: UUID, input_names: List[str]) -> None:
        """Set a notebook cell's input names.

        Args:
            cell_id (UUID): ID of the cell.
            input_names (List[str]): Names for the input variables.
        """
        self.clear_cell(cell_id)

        output_names = set()
        for cell in self.cells:
            if cell.id != cell_id and cell.output_name != "":
                output_names.add(cell.output_name)

        cell = self.get_cell(cell_id)
        for input_name in input_names:
            if input_name not in output_names:
                cell.set_error()
                cell.input_names = []
                raise ValueError(f"No cell with output name \"{input_name}\"")
        cell.input_names = input_names

    def set_cell_code(self, cell_id: UUID, code: str) -> None:
        """Set a notebook cell's code.

        Args:
            cell_id (UUID): ID of the cell.
            code (str): String of code to set on the cell.
        """
        cell = self.get_cell(cell_id)
        cell.code = code
        self.clear_cell(cell.id)

    def get_cell(self, cell_id: UUID) -> Cell:
        """Get a cell from the notebook.

        Args:
            cell_id (UUID): ID of the cell to get from the notebook.
        """
        for cell in self.cells:
            if cell.id == cell_id:
                return cell
        raise ValueError(f"No cell with ID {cell_id} found in notebook")

    def delete_cell(self, cell_id: UUID) -> None:
        """Delete a cell from the notebook.

        Args:
            cell_id (UUID): ID of the cell to delete from the notebook.
        """
        self.cells = [c for c in self.cells if c.id != cell_id]

    def add_cell(self) -> UUID:
        """Add a cell to the notebook.

        Returns:
            UUID: ID of the new cell.
        """
        new_cell = Cell(output_name="")
        self.cells.append(new_cell)
        return new_cell.id

    def _get_children(self, cell: Cell) -> Iterator[Cell]:
        for other in self.cells:
            for input_name in other.input_names:
                if input_name == cell.output_name:
                    yield other

    def _get_parents(self, cell: Cell) -> Iterator[Cell]:
        for other in self.cells:
            for input_name in cell.input_names:
                if other.output_name == input_name:
                    yield other

    def run_cell(self, cell_id: UUID) -> None:
        """Run a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to run.
        """
        cell = self.get_cell(cell_id)

        context: Dict[str, Any] = {}
        for parent in self._get_parents(cell):
            if parent.status == CellStatus.EXPIRED:
                self.run_cell(parent.id)
            context[parent.output_name] = parent.output

        logger.info("Running cell %s", cell.id)
        cell.run(context)

        if cell.status == CellStatus.OK:
            for child in self._get_children(cell):
                self.run_cell(child.id)

    def update_cell_language(self, cell_id: UUID, language: Language) -> None:
        """Run a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to run.
            language (Language): New language to use.
        """
        cell = self.get_cell(cell_id)
        cell.language = language
        cell.clear()

    def clear_cell(self, cell_id: UUID) -> None:
        """Clear a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to clear.
        """
        cell = self.get_cell(cell_id)
        cell.clear()

        for child in self._get_children(cell):
            self.clear_cell(child.id)

    def reorder_cells(self, cell_ids: List[UUID]) -> None:
        """Reorder cells in the notebook.

        Args:
            cell_ids (List[UUID]): List of cell IDs in the correct order.
        """
        new_cells = []
        for cell_id in cell_ids:
            cell = self.get_cell(cell_id)
            new_cells.append(cell)
        self.cells = new_cells

    @classmethod
    def from_filepath(cls, filepath: Path) -> "Notebook":
        """Load a notebook from a .cado notebook file.

        Args:
            filepath (Path): Filepath to a .cado notebook file.

        Returns:
            Notebook: Loaded cado notebook.
        """
        with filepath.open() as f:
            notebook_json = json.load(f)
            version = notebook_json["version"]
            if version != CURRENT_VERSION:
                raise ValueError(f"Cannot load notebook with version {version}")
            return cls.parse_obj(notebook_json)

    def to_filepath(self, filepath: Path) -> None:
        """Save a notebook to a .cado notebook file.

        Args:
            filepath (Path): Filepath to a .cado notebook file.
        """
        with filepath.open("w") as f:
            notebook_json = self.json()
            f.write(notebook_json)

        # TODO: Check that the graph is acyclic



import traceback
from typing import Any, Dict, List, Mapping, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from cado.core.cell_status import CellStatus
from cado.core.language import Language


class Cell(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    code: str = ""

    output_name: str
    output: Optional[Any] = None
    input_names: List[str] = []
    language: Language = Language.PYTHON

    printed: Optional[str] = None
    status: CellStatus = CellStatus.EXPIRED

    def run(self, context: Dict[str, Any]) -> None:
        """Run the cell.

        Returns:
            Any: The output of running the cell.
        """
        if self.code == "":
            self.set_error()
            raise ValueError(f"Code is empty for cell \"{self.id}\"")
        if self.output_name == "":
            self.set_error()
            raise ValueError(f"No output name set for cell \"{self.id}\"")

        exec_locals: Mapping[str, object] = {}
        try:
            # pylint: disable=exec-used
            exec(self.code, context, exec_locals)
        except Exception as exc:
            self.set_error()
            raise ValueError(f"Failed to exec: {traceback.format_exc()}") from exc

        # Check that a variable with the cell output name was emitted by exec
        if self.output_name not in exec_locals:
            self.set_error()
            raise ValueError(f"Cell name \"{self.output_name}\" was not found in exec locals for cell ({self.id})")
        self.output = exec_locals[self.output_name]
        self.status = CellStatus.OK

    def clear(self) -> None:
        """Clear the cell outputs and set the status to expired."""
        self.output = None
        self.printed = None
        self.status = CellStatus.EXPIRED

    def set_error(self) -> None:
        """Clear the cell outputs and set the status to error."""
        self.output = None
        self.status = CellStatus.ERROR



from pathlib import Path
from typing import List

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from cado.app import routes

ALLOWED_ORIGINS: List[str] = []

app = FastAPI()

app.include_router(routes.router)

static_dirpath = Path(__file__).resolve().parent.parent / "ui" / "dist"
app.mount("/", StaticFiles(directory=static_dirpath, html=True), name="static")

app.add_middleware(
    CORSMiddleware,
    allow_origins=ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)



import logging
import traceback
from pathlib import Path
from typing import Any

from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse
from cado.app.example import load_example_notebook

from cado.app.message import (
    ClearCell,
    DeleteCell,
    ErrorResponse,
    GetNotebook,
    GetNotebookResponse,
    Message,
    MessageType,
    NewCell,
    ReorderCells,
    RunCell,
    UpdateCellCode,
    UpdateCellInputNames,
    UpdateCellLanguage,
    UpdateCellOutputName,
)
from cado.core.notebook import Notebook

logger = logging.getLogger(__name__)

router = APIRouter()


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
@router.websocket(path="/stream")
async def stream_api(socket: WebSocket) -> None:
    """Websocket endpoint for streaming commands."""
    logger.info("Starting connection...")

    await socket.accept()
    logger.info("Connection open")

    try:
        # TODO: Make filepath configurable
        filepath = Path("./notebook.cado")
        logger.info("Reading notebook from file: %s", filepath)
        notebook = Notebook.from_filepath(filepath)
    # pylint: disable=broad-exception-caught
    except Exception as exc:
        logger.error("Could not load notebook from file %s: %s", filepath, str(exc))
        example_name = "sum"
        logger.error("Using example notebook %s.cado", example_name)
        notebook = load_example_notebook(example_name)

    try:
        while True:
            message_json = await socket.receive_json()
            try:
                message_type = MessageType.from_str(message_json["type"])
                logger.info("Got message: %s", message_json)

                response = _get_response(message_type, message_json, notebook)
                await socket.send_json(response.json())
            # pylint: disable=broad-exception-caught
            except Exception as exc:
                logger.error("Exception raised during session loop")
                logger.error("Traceback: %s", traceback.format_exc())
                response = ErrorResponse(error=str(exc))
                await socket.send_json(response.json())
    except WebSocketDisconnect:
        logger.info("Websocket disconnected")
        # TODO: Make filepath configurable
        filepath = Path("./notebook.cado")
        logger.info("Writing notebook to file: %s", filepath)
        notebook.to_filepath(filepath)


def _get_response(message_type: MessageType, message_json: Any, notebook: Notebook) -> Message:
    if message_type == MessageType.GET_NOTEBOOK:
        GetNotebook.parse_obj(message_json)
    elif message_type == MessageType.UPDATE_CELL_CODE:
        update_cell_code = UpdateCellCode.parse_obj(message_json)
        notebook.set_cell_code(update_cell_code.cell_id, update_cell_code.code)
    elif message_type == MessageType.UPDATE_CELL_OUTPUT_NAME:
        update_cell_output_name = UpdateCellOutputName.parse_obj(message_json)
        notebook.update_cell_output_name(update_cell_output_name.cell_id, update_cell_output_name.output_name)
    elif message_type == MessageType.RUN_CELL:
        run_cell = RunCell.parse_obj(message_json)
        notebook.run_cell(run_cell.cell_id)
    elif message_type == MessageType.CLEAR_CELL:
        clear_cell = ClearCell.parse_obj(message_json)
        notebook.clear_cell(clear_cell.cell_id)
    elif message_type == MessageType.NEW_CELL:
        NewCell.parse_obj(message_json)
        notebook.add_cell()
    elif message_type == MessageType.DELETE_CELL:
        delete_cell = DeleteCell.parse_obj(message_json)
        notebook.delete_cell(delete_cell.cell_id)
    elif message_type == MessageType.UPDATE_CELL_INPUT_NAMES:
        update_cell_input_names = UpdateCellInputNames.parse_obj(message_json)
        notebook.update_cell_input_names(update_cell_input_names.cell_id, update_cell_input_names.input_names)
        return GetNotebookResponse(notebook=notebook)
    elif message_type == MessageType.UPDATE_CELL_LANGUAGE:
        update_cell_language = UpdateCellLanguage.parse_obj(message_json)
        notebook.update_cell_language(update_cell_language.cell_id, update_cell_language.language)
        return GetNotebookResponse(notebook=notebook)
    elif message_type == MessageType.REORDER_CELLS:
        reorder_cells = ReorderCells.parse_obj(message_json)
        notebook.reorder_cells(reorder_cells.cell_ids)
        return GetNotebookResponse(notebook=notebook)
    else:
        logger.error("Request type did not match any known message types")
        return ErrorResponse(error=f"Unknown message type from client: {message_type}")
    return GetNotebookResponse(notebook=notebook)


@router.get(path="/status")
def get_status() -> JSONResponse:
    """Endpoint for app status."""
    logger.info("GET app status")
    return JSONResponse({
        'status': 'healthy',
    })



from enum import Enum
from typing import List
from uuid import UUID

from pydantic import BaseModel

from cado.core.cell import Cell
from cado.core.language import Language
from cado.core.notebook import Notebook


class MessageType(Enum):
    # publish
    GET_NOTEBOOK = "get-notebook"
    UPDATE_CELL_CODE = "update-cell-code"
    UPDATE_CELL_OUTPUT_NAME = "update-cell-output-name"
    UPDATE_CELL_INPUT_NAMES = "update-cell-input-names"
    UPDATE_CELL_LANGUAGE = "update-cell-language"
    RUN_CELL = "run-cell"
    CLEAR_CELL = "clear-cell"
    NEW_CELL = "new-cell"
    DELETE_CELL = "delete-cell"
    REORDER_CELLS = "reorder-cells"

    # subscribe
    GET_NOTEBOOK_RESPONSE = "get-notebook-response"
    GET_CELL_RESPONSE = "get-cell-response"
    ERROR_RESPONSE = "error-response"

    @classmethod
    def from_str(cls, message_name: str) -> 'MessageType':
        """Parse MessageType from string.

        Args:
            message_name (str): Name of message type to parse.

        Raises:
            ValueError: If the name was not a valid message type.

        Returns:
            MessageType: The parsed MessageType.
        """
        for _, enum_value in cls.__members__.items():
            if enum_value.value == message_name:
                return enum_value
        raise ValueError(f"Unknown message type '{message_name}'")


class Message(BaseModel):
    type: MessageType


# region: publish


class GetNotebook(Message):
    type: MessageType = MessageType.GET_NOTEBOOK


class UpdateCellCode(Message):
    cell_id: UUID
    code: str
    type: MessageType = MessageType.UPDATE_CELL_CODE


class UpdateCellOutputName(Message):
    cell_id: UUID
    output_name: str
    type: MessageType = MessageType.UPDATE_CELL_OUTPUT_NAME


class UpdateCellInputNames(Message):
    cell_id: UUID
    input_names: List[str]
    type: MessageType = MessageType.UPDATE_CELL_INPUT_NAMES


class UpdateCellLanguage(Message):
    cell_id: UUID
    language: Language
    type: MessageType = MessageType.UPDATE_CELL_LANGUAGE


class RunCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.RUN_CELL


class ClearCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.CLEAR_CELL


class NewCell(Message):
    type: MessageType = MessageType.NEW_CELL


class DeleteCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.DELETE_CELL


class ReorderCells(Message):
    cell_ids: List[UUID]
    type: MessageType = MessageType.REORDER_CELLS


# endregion: publish

# region: subscribe


class GetNotebookResponse(Message):
    notebook: Notebook
    type: MessageType = MessageType.GET_NOTEBOOK_RESPONSE


class GetCellResponse(Message):
    cell: Cell
    type: MessageType = MessageType.GET_CELL_RESPONSE


class ErrorResponse(Message):
    error: str
    type: MessageType = MessageType.ERROR_RESPONSE


# endregion: subscribe



import logging
from pathlib import Path

import click
import uvicorn

from cado.app.app import app as cado_app


@click.group()
def main() -> None:
    """Main CLI entrypoint."""


@main.command(name="up")
@click.option("--host", type=str, default="0.0.0.0")
@click.option("--port", "-p", type=int, default=8000)
@click.option("--debug", is_flag=True)
@click.option("--log-level", type=int, default=logging.INFO)
def up_command(host: str, port: int, debug: bool, log_level: int) -> None:
    """Command to start up cado app."""
    print("Welcome to cado IDE")
    current_dirpath = Path(__file__).parent
    print(f"From {current_dirpath}")
    print("Starting app")

    package_dirpath = current_dirpath.parent.parent
    log_config_filepath = package_dirpath / "logging.yaml"
    uvicorn.run(
        cado_app,
        host=host,
        port=port,
        debug=debug,
        log_level=log_level,
        log_config=str(log_config_filepath),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cado', 'cado.app', 'cado.cli', 'cado.core', 'cado.examples']

package_data = \
{'': ['*'],
 'cado': ['ui/*',
          'ui/dist/*',
          'ui/dist/assets/*',
          'ui/public/*',
          'ui/src/*',
          'ui/src/components/*',
          'ui/src/hooks/*',
          'ui/src/images/*',
          'ui/src/lib/*',
          'ui/src/lib/models/*',
          'ui/src/widgets/*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'fastapi>=0.70.0,<0.71.0',
 'pydantic>=1.9.0,<2.0.0',
 'uvicorn[standard]>=0.15.0,<0.16.0']

entry_points = \
{'console_scripts': ['cado = cado.cli.cli:main']}

setup_kwargs = {
    'name': 'cado',
    'version': '0.1.0rc2',
    'description': 'Python notebook development environment.',
    'long_description': '<div align="center">\n  <img src="assets/cado-banner.png">\n  <h1>cado</h1>\n\n  <p>\n    <strong>Python notebook IDE with a focus on reactivity</strong>\n  </p>\n\n  <br>\n  <div>\n    <a href="https://badge.fury.io/py/cado"><img src="https://badge.fury.io/py/cado.svg" alt="PyPI"></a>\n    <a href="https://pepy.tech/project/cado"><img src="https://pepy.tech/badge/cado" alt="Downloads"></a>\n    <a href="https://github.com/gregorybchris/cado/actions/workflows/ci.yaml"><img src="https://github.com/gregorybchris/cado/actions/workflows/ci.yaml/badge.svg" alt="CI"></a>\n  </div>\n  <br>\n</div>\n\n## About\n\n`cado` is a notebook IDE for Python, like [Jupyter](https://jupyter.org/), but with a reactive cell model, taking inspiration from [Observable](https://observablehq.com/). Each cell defines its own outputs that other cells can listen to. When the output of a parent cell updates, the change propagates to all child cells. When a child runs, it uses cached outputs from parent cells, reducing the amount of computation per cell.\n\n## Installation\n\n```bash\npip install cado\n```\n\n## Usage\n\n```bash\n# Start up a cado server\ncado up\n```\n',
    'author': 'Chris Gregory',
    'author_email': 'christopher.b.gregory@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gregorybchris/cado',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4',
}


setup(**setup_kwargs)



from pathlib import Path

from cado.core.notebook import Notebook


def load_example_notebook(example_name: str) -> Notebook:
    """Load a simple example notebook.

    Returns:
        Notebook: A notebook containing a single example cell.
    """
    examples_dirpath = Path(__file__).parent.parent / "examples"
    notebook_filepath = examples_dirpath / f"{example_name}.cado"
    return Notebook.from_filepath(notebook_filepath)
